/*
 * Copyright 1992 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Kevin S Braunsdorf, ksb@cc.purdue.edu, purdue!ksb
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

/*
 * this is the configuration file for the Ohio State/PUCC console
 * server.  Just define the macros below to somehting that looks good
 * and give it a go.  It'll complain (under conserver -V) if things
 * look really bad.
 */
/* #define IBMR2 1	/**/
/* #define HPUX7 1	/**/
/* #define SUN4 1	/**/

/* some machine specific details
 */
#if !defined(USE_OLDSEL)
#if	defined(IBMR2)
#include <sys/select.h>
#endif
#endif
#if !defined(HAVE_UWAIT)
#define HAVE_UWAIT	!defined(IBMR2)
#endif

/* This is the port number used in the connection.  It can use either
 * /etc/services or a hardcoded port (SERVICE name has precedence).
 */
/* #define PORT 2015 */
#define SERVICE		"conserver"

/* The name of the host which will act as the console server
 */
#define HOST		"localhost"

/* the default escape sequence used to give meta commands
 */
#define DEFATTN		'\005'
#define DEFESC		'c'

/* Location of the configuration file
 */
#define CONFIG		"/usr/local/lib/conserver.cf"


/* The maximum number of serial lines that can be handled by a child process
 */
#define MAXMEMB		10


/* The maximum number of child processes spawned.
 */
#define MAXGRP		32

/* the max number of characters conserver will replay for you (the r command)
 */
#define MAXREPLAY	(80*25)

/* the console server will provide a pseudo-device console which
 * allows operators to run backups and such without a hard wired
 * line (this is also good for testing the server to see if you
 * might wanna use it).  Turn this on only if you need it.
 */
#define DO_VIRTUAL	1

#if DO_VIRTUAL
/* if the virtual console option is on we need a source to ptys,
 * the PUCC ptyd daemon is the best source be know, else fall back
 * on some emulation code?? (XXX)
 */
#if !defined(HAVE_PTYD)
#define HAVE_PTYD	!(defined(IBMR2)||defined(NEXT2)||defined(ETA10)||defined(V386))
#endif
#if !defined(HAVE_SETSID)
#define HAVE_SETSID	(defined(IBMR2)||defined(HPUX7))
#endif
#endif

/* should we use flock to keep multiple conservers from hurting each other?
 */
#if !defined(USE_FLOCK)
#define USE_FLOCK	!defined(IBMR2)
#endif

/* should we try to pop streams modules off?
 */
#if !defined(USE_STRREAMS)
#define USE_STREAMS	defined(SUN4)
#endif

/* if we do not have old style tty emulation use termios.h (HPUX7 only now)
 */
#if !defined(USE_TERMIO)
#define USE_TERMIO	(defined(ETA10)||defined(V386))
#endif
#if !defined(USE_TERMIOS)
#define USE_TERMIOS	defined(HPUX7)
#endif

/* used to force the server process to clear parity, which is for farmers
 */
#define CPARITY		1


/* if you do not have fd_set's here is a possible emulation
 */
#if USE_OLDSEL
typedef long fd_set;

#define FD_ZERO(a) {*(a)=0;}
#define FD_SET(d,a) {*(a) |= (1 << (d));}
#define FD_CLR(d,a) {*(a) &= ~(1 << (d));}
#define FD_ISSET(d,a) (*(a) & (1 << (d)))
#endif

#if USE_TERMIOS
#define TCGETS  _IOR('T', 16, struct termios)
#define TCSETS  _IOW('T', 17, struct termios)
#endif

/* which type does wait(2) take for status location
 */
#if HAVE_UWAIT
#define WAIT_T	union wait
#if ! defined WEXITSTATUS
#define WEXITSTATUS(x)	((x).w_retcode)
#endif
#else
#define WAIT_T	int
#endif

/* which type signal handlers return on this machine
 */
#if defined(sun) || defined(NEXT2)
#define SIGRETS	void
#else
#define SIGRETS	int
#endif

/* do we have a (working) setsockopt call
 */
#if !defined(HAVE_SETSOCKOPT)
#define HAVE_SETSOCKOPT	(defined(sun))
#endif

/* does this system have the ANSI strerror() function?
 */
#define HAVE_STRERROR	(defined(IBMR2)||defined(ETA10)||defined(V386))
#if ! HAVE_STRERROR
extern int errno;
extern char *sys_errlist[];
#define strerror(Me) (sys_errlist[Me])
#endif

/* that's all.  just run
 *	make
 *	./conserver -V
 */

/* communication constants
 */
#define OB_SUSP		'Z'		/* suspended by server		*/
#define OB_DROP		'.'		/* dropped by server		*/

/* Due to C's poor man's macros the macro below would break if statements,
 * What we want
 *	macro()			{ stuff }
 * but the syntax gives us
 *	macro()			{ stuff };
 *
 * the extra semicolon breaks if statements!
 * Of course, the one we use makes lint scream:
 *	macro()			do { stuff } while (0)
 *
 * which is a statement and makes if statements safe
 */
#if defined(lint)
extern int shut_up_lint;
#else
#define shut_up_lint	0
#endif

/* this macro efficently outputs a constant string to a fd
 * of course it doesn't check the write :-(
 */
#define CSTROUT(Mfd, Mstr)	do {	\
	static char _ac[] = Mstr; \
	write(Mfd, _ac, sizeof(_ac)-1); \
	} while (shut_up_lint)
