.\"@Explode @"\.
.\" $Id: srtunq.man,v 5.1 1997/01/09 03:32:12 kb207252 Exp $
.\"	# my routine titler -- accepts the lines of the title as args
.de TM
.in .5i
.B \\$1
.br
.B \\$2
.br
.B \\$3
.br
.B \\$4
.br
.B \\$5
.br
.B \\$6
.br
.B \\$7
.br
.B \\$8
.br
.B \\$9
.br
.ft R
.in +.5i
..
'
'
'
.TH SRTUNQ 3L PUCC
.SH NAME
srtdel, srtinit, srtin, srtmem, srtgti, srtgets, srtapply, srtfree, srtdtree \- in memory
string sorting
.SH SYNOPSIS
.B #include <stdio.h>
.br
.B #include \*(lqsrtunq.h\*(rq
.sp
cc \-I/usr/include/local file.c
.B \-lsrtunq
.SH DESCRIPTION
.I Libsrtunq.a
is used to extract unique items from a possibly long list,
where items are likely to be replicated numerously.
The list of unique items must be small enough to
fit in memory, but the number of repetitions is possibly high.
.PP
The caller has control over the database through the use of a
\fBSRTTABLE\fP variable.  The subroutines provide for data entry and
retrieval, memory allocation and deallocation.
.SH ROUTINES
.TM "typedef  struct ... SRTTABLE;"
Users will define one variable of this type to hold the root of
each sorted list they wish to keep.
Most routines in this library require the address of such
a variable as their first argument.

.TM "typedef  struct ... SRTENTRY;"
Strings are kept in a structure of this type internally.
The user should not depend on the internal details of this type.

.TM "void" "srtinit(tbl)" "SRTTABLE *tbl;"
This subroutine must be called to initialize the database tag \fItbl\fP
before any data are entered or retrieved from that tree.
It assumes that the tag has not been used
to store a tree, and therefore does not attempt to free any such data.

.TM "char *" "srtin(tbl, string, compare)" "SRTTABLE *tbl;" "char *string;" "int (*compare)();"
The existing data tree is searched for the string,
if it is found then a pointer to that string is returned.
Otherwise, space is allocated for the string and pointer structure via
\fImalloc\fP(3).
The string is copied to this new space which is linked into the tree
and a pointer to the new string is returned.
If space cannot be obtained, the operation is aborted and NULL is returned
(the data structure remains consistent, but the string is not added).
The strings are compared and sorted with the subroutine pointed to by
\fIcompare\fP.  This subroutine takes two string pointers as arguments.
It returns zero if the strings are the same,
less than zero if the first string should precede the second, and
greater than zero if the second string should precede the first.
Use \fIstrcmp\fP(3) if simple lexicographical ordering is desired.
It is confusing at best if different \fIcompare\fP functions are used
when inserting strings into a given tree.

.TM "char *" "srtmem(tbl, string, compare)" "SRTTABLE *tbl;" "char *string;" "int (*compare)();"
Return the database entry for \fIstring\fP if it is already a member of
the table, else NULL.

.TM "int" "srtdel(tbl, string, compare)" "SRTTABLE *tbl;" "char *string;" "int (*compare)();"
The existing data tree is searched for the string,
if it is found then it is deleted and a nonzero value is returned.
Otherwise, 0 is returned.

.TM "void" "srtgti(tbl);" "SRTTABLE *tbl;"
This subroutine initializes the database tag pointed to by \fItbl\fP
so that a tree traversal can be made via \fIsrtgets\fP.

.TM "char *" "srtgets(tbl);" "SRTTABLE *tbl;"
This subroutine extracts the next string from the data structure.
The strings are returned in the order specified by the \fIcompare\fP
function when they were inserted with \fIsrtin\fP.
When the list is exhausted, NULL is returned.

.TM "int" "srtapply(tbl, func)" "SRTTABLE *tbl;" "int (*func)();"
This subroutine applies the \fIfunc\fP to each string in the tree (order
determined by \fIcompare\fP when they were inserted) until the func returns
non-zero, or there are no more strings.  If the \fIfunc\fP returned
non-zero then that value is returned, otherwise 0 is returned.

.TM "void" "srtfree(tbl)" "SRTTABLE *tbl;"
This subroutine deletes a database, and re-initializes the
database tag.
It assumes that the database tag was initialized at one time via \fIsrtinit\fP
(other routines will probably also have been called).
The space formerly occupied by string data and pointer structures is
deallocated via \fIfree\fP(3).

.TM "void" "srtdtree(tbl, ent)" "SRTTABLE *tbl;" "SRTENTRY *ent;"
This subroutine recursively deletes a database subtree.
The space formerly occupied by the string data and pointer structures is
deallocated via \fIfree\fP(3).
This routine is most likely only of use internally.
.SH EXAMPLE
.nf
#include <stdio.h>
main()
{
	extern int strcmp();
	SRTTABLE tree;
	char buf[80], *p;
	int i;

	/* init the tree */
	srtinit(&tree);

	/* add some strings */
	while (NULL != fgets(buf, 80, stdin))  /* want the \en terminator */
		if (NULL == (p = srtin(&tree, buf, strcmp)))
			printf("out of memory!\en");
	
	/* init tree for srtgets */
	srtgti(&tree);

	/* print out the strings with srtapply and printf -- can't have
	 * a legal printf % escape in the strings! */
	(void) srtapply(&tree, printf);

	/* use srtgets to print the strings out this time -- keep count */
	for (i = 0; NULL != (p = srtgets(&tree)); ++i)
		printf("string %2d is: %s\en", i, p);
	printf("there were %d strings\en", i);

	/* free the database */
	srtfree(&tree);
}
.fi
.SH DIAGNOSTICS
There are no messages printed by these routines.
Catchable errors are returned as NULL.
Compiled in errors such as the use of strings that
are not null terminated tend to result in core files.
.SH FILES
/usr/local/lib/libsrtunq.a
.br
/usr/include/local/srtunq.h
.SH SEE ALSO
malloc(3), free(3), qsort(3), strcmp(3)
.SH AUTHOR
Stephen Uitti, PUCC
.SH BUGS
The structure names and typedefs seem confused.  The typedefs make more sense.
